
def parse(exc):
    """Takes in an exception and returns the boto
    `Message, Code` properties if they exist - else `None, None`
    """
    if exc is not None and hasattr(exc, 'response') and exc.response is not None:
      error = exc.response.get('Error', {})
      code = error.get('Code', None)
      message = error.get('Message', None)
      return message, code
    return None, None

class DynamofException(Exception):
   def __init__(self, message):
       self.message = message
       super().__init__(message)
   def info(self, add_message):
       """ Adds details to the base error message"""
       self.message = f'{self.message}; {add_message}'
       return self

class PreexistingTableException(DynamofException):
    def __init__(self):
        message = "Attempted to create a table that already exists"
        super().__init__(message)
    @classmethod
    def matches(cls, err):
        message, _ = parse(err)
        key = 'Cannot create preexisting table'
        return message == key

class TableDoesNotExistException(DynamofException):
    def __init__(self):
        message = "Attempted to do operation on a table that does not exist"
        super().__init__(message)
    @classmethod
    def matches(cls, err):
        message, _ = parse(err)
        key = 'Cannot do operations on a non-existent table'
        return message == key

class ConditionNotMetException(DynamofException):
    def __init__(self):
        message = "Could not find an item that satisifed the given conditions"
        super().__init__(message)
    @classmethod
    def matches(cls, err):
        _, code = parse(err)
        return code == 'ConditionalCheckFailedException'

class BadGatewayException(DynamofException):
    def __init__(self):
        message = "Issue communicating with dynamo"
        super().__init__(message)
    @classmethod
    def matches(cls, err):
        message, _ = parse(err)
        return message == 'Bad Gateway'

class UnknownDatabaseException(DynamofException):
    def __init__(self):
        message = "An unkonwn exception occured when executing request to dynamo"
        super().__init__(message)



import decimal
from boto3.dynamodb.types import TypeSerializer, TypeDeserializer


def update_expression(data):
    """Builds a string thats a dynamo friendly
    expression"""
    keys = [f'{key} = :{key}' for key, value in data.items()]
    key_expression = ', '.join(keys)
    return f'SET {key_expression}'

def expression_attribute_values(data):
    """Builds an object that contains ':'
    prepended to the key to make a dynamo
    friendly ExpressionAttributeValues"""
    values = value_type_tree(data)
    exp_values = {}
    for key, value in values.items():
        exp_values[f':{key}'] = value
    return exp_values


def condition_expression(id):
    key = 'id' # default
    if isinstance(id, dict):
        keys = list(id.keys())
        key = keys[0]
    return f'{key} = :{key}'

def key_schema(hash_key):
    return [{
        'AttributeName': hash_key,
        'KeyType': 'HASH'
    }]

def attribute_definitions(keys):
    return [{
        'AttributeName': key,
        'AttributeType': 'S'
    } for key in keys]

def provisioned_throughput():
    return {
        'ReadCapacityUnits': 1,
        'WriteCapacityUnits': 1
    }

def deep_strip_Decimals(obj):
    """Patches an issue with dynamo where it takes in number
    types but always returns Decimal class type.
    https://github.com/boto/boto3/issues/369
    """
    if isinstance(obj, list):
        return [deep_strip_Decimals(item) for item in obj]
    if isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = deep_strip_Decimals(v)
        return obj
    if isinstance(obj, decimal.Decimal):
        if obj % 1 == 0:
            return int(obj)
        return float(obj)
    return obj

def value_type_tree(data):

    if data is None:
        return None

    s = TypeSerializer()

    result = {}
    for key, val in data.items():
        result[key] = s.serialize(val)

    return result

def destructure_type_tree(data):

    if data is None:
        return None

    d = TypeDeserializer()

    result = {}

    for key, val in data.items():
        result[key] = d.deserialize(val)

    return deep_strip_Decimals(result)



import collections
from dynamof import builder as ab

Attribute = collections.namedtuple("Attribute", [
    "equals",
    "gt",
    "lt"
    # "lt_or_eq",
    # "gt_or_eq"
])

Condition = collections.namedtuple("Condition", [
    "expression",
    "attr_values"
])

# condition and
def cand(*conditions):
    exp = ' AND '.join([cond.expression for cond in conditions])
    values = collections.ChainMap(*[cond.attr_values for cond in conditions])
    return Condition(exp, values)

# condition or
def cor(*conditions):
    exp = ' OR '.join([cond.expression for cond in conditions])
    values = collections.ChainMap(*[cond.attr_values for cond in conditions])
    return Condition(exp, values)


def attr(name):

    def equals(value):
        exp = f'{name} = :{name}'
        values = ab.value_type_tree({ f':{name}': value })
        return Condition(exp, values)

    def greater_than(value):
        exp = f'{name} > :{name}'
        values = ab.value_type_tree({ f':{name}': value })
        return Condition(exp, values)

    def less_than(value):
        exp = f'{name} < :{name}'
        values = ab.value_type_tree({ f':{name}': value })
        return Condition(exp, values)

    return Attribute(equals, greater_than, less_than)




from botocore.exceptions import ClientError

from dynamof import response
from dynamof.exceptions import (
    UnknownDatabaseException,
    PreexistingTableException,
    ConditionNotMetException,
    BadGatewayException,
    TableDoesNotExistException
)

def create(allow_existing):
    def run(client, description):
        try:
            res = client.create_table(**description)
        except Exception as err:  # pylint: disable=broad-except
            if BadGatewayException.matches(err): raise BadGatewayException()
            if PreexistingTableException.matches(err):
                if allow_existing is False:
                    table_name = description.get('TableName')
                    raise PreexistingTableException().info(f'table={table_name}')
                return response.create_response(None, skipped=True)
            raise UnknownDatabaseException()
        return response.create_response(res)

    return run

def find():
    def run(client, description):
        try:
            res = client.get_item(**description)
        except ClientError as err:
            if BadGatewayException.matches(err): raise BadGatewayException()
            if TableDoesNotExistException.matches(err): raise TableDoesNotExistException().info(description.get('TableName'))
            if ConditionNotMetException.matches(err): raise ConditionNotMetException().info(description.get('ConditionExpression'))
            raise UnknownDatabaseException()
        return response.find_response(res)
    return run

def add():
    def run(client, description):
        try:
            res = client.put_item(**description)
        except ClientError as err:
            if BadGatewayException.matches(err): raise BadGatewayException()
            if TableDoesNotExistException.matches(err): raise TableDoesNotExistException().info(description.get('TableName'))
            raise UnknownDatabaseException()
        return response.add_response(res)
    return run

def update():
    def run(client, description):
        try:
            res = client.update_item(**description)
        except ClientError as err:
            if BadGatewayException.matches(err): raise BadGatewayException()
            if TableDoesNotExistException.matches(err): raise TableDoesNotExistException().info(description.get('TableName'))
            if ConditionNotMetException.matches(err): raise ConditionNotMetException().info(description.get('ConditionExpression'))
            raise UnknownDatabaseException()
        return response.update_response(res)
    return run

def delete():
    def run(client, description):
        try:
            res = client.delete_item(**description)
        except ClientError as err:
            if BadGatewayException.matches(err): raise BadGatewayException()
            if TableDoesNotExistException.matches(err): raise TableDoesNotExistException().info(description.get('TableName'))
            if ConditionNotMetException.matches(err): raise ConditionNotMetException().info(description.get('ConditionExpression'))
            raise UnknownDatabaseException()
        return response.delete_response(res)
    return run

def query():
    def run(client, description):
        try:
            res = client.query(**description)
        except ClientError as err:
            if BadGatewayException.matches(err): raise BadGatewayException()
            if TableDoesNotExistException.matches(err): raise TableDoesNotExistException().info(description.get('TableName'))
            if ConditionNotMetException.matches(err): raise ConditionNotMetException().info(description.get('ConditionExpression'))
            raise UnknownDatabaseException()
        return response.query_response(res)
    return run



from dynamof import builder as ab


get_data = lambda type_tree: ab.destructure_type_tree(type_tree)
get_retries = lambda response: response.get('ResponseMetadata', {}).get('RetryAttempts', None)
get_success = lambda response: response.get('ResponseMetadata', {}).get('HTTPStatusCode', 0) == 200
get_count = lambda response: response.get('Count', None)
get_scanned_count = lambda response: response.get('ScannedCount', None)


def create_response(response, skipped=False):

    return {
        'table_already_existed': response is None and skipped is True,
        'retries': get_retries(response or {}),
        'success': get_success(response or {}),
        'raw': response
    }


def find_response(response):

    return {
        'item': get_data(response.get('Item', {})),
        'retries': get_retries(response),
        'success': get_success(response),
        'raw': response
    }


def add_response(response):

    return {
        'item': get_data(response.get('Attributes', {})),
        'retries': get_retries(response),
        'success': get_success(response),
        'raw': response
    }


def update_response(response):

    return {
        'item': get_data(response.get('Attributes', {})),
        'retries': get_retries(response),
        'success': get_success(response),
        'raw': response
    }


def query_response(response):

    return {
        'items': [get_data(type_tree) for type_tree in response.get('Items', [])],
        'count': get_count(response),
        'scanned_count': get_scanned_count(response),
        'retries': get_retries(response),
        'success': get_success(response),
        'raw': response
    }


def delete_response(response):

    return {
        'retries': get_retries(response),
        'success': get_success(response),
        'raw': response
    }



import copy
import collections
from dynamof import builder as ab
from dynamof import runners
from dynamof.utils import new_id


Operation = collections.namedtuple("Operation", [
    "description",
    "runner"
])


def create(table_name, hash_key, allow_existing=False):
    description = dict(
        TableName=table_name,
        KeySchema=ab.key_schema(hash_key),
        AttributeDefinitions=ab.attribute_definitions([hash_key]),
        ProvisionedThroughput=ab.provisioned_throughput()
    )
    return Operation(description, runners.create(
        allow_existing=allow_existing
    ))


def find(table_name, key):
    description = dict(
        TableName=table_name,
        Key=ab.value_type_tree(key)
    )
    return Operation(description, runners.find())


def add(table_name, item, auto_inc=False):
    attributes = copy.deepcopy(item)
    if auto_inc is True:
        attributes['id'] = new_id()
    description = dict(
        TableName=table_name,
        Item=ab.value_type_tree(attributes),
        ReturnValues='ALL_OLD'
    )
    return Operation(description, runners.add())


def update(table_name, key, attributes):
    description = dict(
        TableName=table_name,
        Key=ab.value_type_tree(key),
        ConditionExpression=ab.condition_expression(key),
        UpdateExpression=ab.update_expression(attributes),
        ExpressionAttributeValues=ab.expression_attribute_values({**key, **attributes}),
        ReturnValues='ALL_NEW'
    )
    return Operation(description, runners.update())


def delete(table_name, key):
    description = dict(
        TableName=table_name,
        Key=ab.value_type_tree(key),
        ConditionExpression=ab.condition_expression(key),
        ExpressionAttributeValues=ab.expression_attribute_values(key)
    )
    return Operation(description, runners.delete())

def query(table_name, conditions):
    description = dict(
        TableName=table_name,
        KeyConditionExpression=conditions.expression,
        ExpressionAttributeValues=conditions.attr_values
    )
    return Operation(description, runners.query())



import setuptools

VERSION='1.0.5'

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dynamof",
    version=VERSION,
    author="Ray Epps",
    author_email="rayharryepps@gmail.com",
    description="A small interface for more easily making calls to dynamo using boto.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rayepps/dynamof",
    download_url=f'https://github.com/rayepps/dynamof/archive/{VERSION}.tar.gz',
    packages=setuptools.find_packages(exclude="test"),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent"
    ],
    python_requires='>=3',
    install_requires=[
        'boto3',
        'botocore'
    ]
)
